import asyncio
import os

from test import AioDiskDBTestCase, run_test_db


class AioDBTestExitSignals(AioDiskDBTestCase):
    def setUp(self, *a, **kw):
        super().setUp(max_file_size=1, max_buffer_size=1)
        self._added_location = None

    async def _persist_data(self):
        while not self.sut.running:
            await asyncio.sleep(1)
        self._added_location = await self.sut.add(b'data')

    @run_test_db
    async def test(self):
        """
        test that the stop signals hook writes data before exiting
        """
        await self._persist_data()
        self.assertIsNotNone(self._added_location)
        with self.assertRaises(FileNotFoundError):
            os.path.getsize(self._path + '/data00000.dat')
        with self.assertRaises(SystemExit):
            self.sut.on_stop_signal()
        self.sut.on_stop_signal()  # fire it twice, it must be executed once
        self.assertEqual(
            self.sut._file_header_size + 4,
            os.path.getsize(self._path + '/data00000.dat')
        )
        self._setup_sut()
        asyncio.get_event_loop().create_task(self.sut.run())
        while not self.sut.running:
            await asyncio.sleep(1)
        data = await self.sut.read(self._added_location)
        self.assertEqual(b'data', data)



class AioDiskDBException(Exception):
    pass


class RunningException(AioDiskDBException):
    pass


class NotRunningException(AioDiskDBException):
    pass


class TimeoutException(AioDiskDBException):
    pass


class ReadTimeoutException(AioDiskDBException):
    pass


class DBNotInitializedException(AioDiskDBException):
    pass


class InvalidDataFileException(AioDiskDBException):
    pass


class ReadOnlyDatabaseException(AioDiskDBException):
    pass


class FilesInconsistencyException(AioDiskDBException):
    pass


class WriteFailedException(AioDiskDBException):
    pass


class InvalidConfigurationException(AioDiskDBException):
    pass


class EmptyTransactionException(AioDiskDBException):
    pass


class TransactionCommitOnGoingException(AioDiskDBException):
    pass


class TransactionAlreadyCommittedException(AioDiskDBException):
    pass


class InvalidDBStateException(AioDiskDBException):
    pass


class IndexDoesNotExist(AioDiskDBException):
    pass


class EmptyPayloadException(AioDiskDBException):
    pass


class InvalidTrimCommandException(AioDiskDBException):
    pass



import os
from aiodiskdb import exceptions
from aiodiskdb.local_types import ItemLocation, WriteEvent
from test import AioDiskDBTestCase, run_test_db


class TestRTRIM(AioDiskDBTestCase):
    @run_test_db
    async def test(self):
        item_location = await self.sut.add(b'test_1')
        self.assertEqual(
            ItemLocation(0, 0, 6),
            item_location
        )
        item_location_2 = await self.sut.add(b'test_2')
        self.assertEqual(
            ItemLocation(0, 6, 6),
            item_location_2
        )
        await self.sut.stop()  # stop the sut, ensures the data write
        self.assertEqual(
            WriteEvent(index=0, position=0, size=12),
            self._writes[0][1]
        )
        self._setup_sut()  # re-instance the sut from scratch.
        await self._run()
        read1 = await self.sut.read(item_location_2)
        self.assertEqual(b'test_2', read1)
        read2 = await self.sut.read(item_location_2)
        self.assertEqual(b'test_2', read2)
        slice_size = await self.sut.rtrim(0, 9, safety_check=b't_2')
        slice_size_2 = await self.sut.rtrim(0, 6)
        self.assertEqual(slice_size + slice_size_2, len(read2))
        self.assertIsNone(await self.sut.read(item_location_2))
        item_location_2 = await self.sut.add(b'test_3')
        self.assertEqual(
            ItemLocation(0, 6, 6),
            item_location_2
        )
        read3 = await self.sut.read(item_location_2)
        self.assertEqual(b'test_3', read3)
        self.sut.disable_overwrite()
        with self.assertRaises(exceptions.ReadOnlyDatabaseException):
            await self.sut.rtrim(0, 6)
        self.sut.enable_overwrite()


class TestTrimIndexDoesNotExist(AioDiskDBTestCase):
    @run_test_db
    async def test(self):
        with self.assertRaises(exceptions.IndexDoesNotExist):
            await self.sut.rtrim(99, 111)

        with self.assertRaises(exceptions.IndexDoesNotExist):
            await self.sut.rtrim(-1, 111)


class TestTrimWholeFile(AioDiskDBTestCase):
    @run_test_db
    async def test(self):
        for _ in range(0, 20):
            await self.sut.add(os.urandom(1024**2))
        await self.sut.rtrim(0, 0)
        with self.assertRaises(FileNotFoundError):
            os.path.getsize(self._path + '/data00000.dat')
        for _ in range(0, 20):
            await self.sut.add(os.urandom(1024**2))
        with self.assertRaises(exceptions.InvalidTrimCommandException):
            await self.sut.rtrim(0, 0, safety_check=b'wrong_bytes')
        with self.assertRaises(exceptions.InvalidTrimCommandException):
            await self.sut.rtrim(0, 1, safety_check=b'wrong_bytes')



import os
import shutil

from aiodiskdb import exceptions
from test import AioDiskDBTestCase


class TestCheckpoints(AioDiskDBTestCase):
    def setUp(
            self, *a,
            max_file_size=16,
            max_buffer_size=1
    ):
        super().setUp(*a, max_file_size=max_file_size, max_buffer_size=max_buffer_size)

    async def test(self):
        await self._run()

        with self.assertRaises(exceptions.InvalidDBStateException):
            await self.sut._clean_db_checkpoint(333)

        checkpoint_id = 333
        data = os.urandom(1024**2)
        data2 = b'daf'
        item_location = await self.sut.add(data)
        await self.sut._flush_buffer()
        await self.sut._write_db_checkpoint(checkpoint_id, 0)
        shutil.move(self._path + '/.checkpoint-333', self._path + '/_checkpoint')
        second_location_add = await self.sut.add(data2)
        assert await self.sut.read(second_location_add) == data2
        await self._stop()
        shutil.move(self._path + '/_checkpoint', self._path + '/.checkpoint-333')

        with self.assertRaises(exceptions.InvalidDBStateException):
            self._setup_sut(clean_stale_data=False)
        self._setup_sut()

        with self.assertRaises(FileNotFoundError):
            os.path.getsize(self._path + '/.checkpoint-333')

        await self._run()
        self.assertIsNone(await self.sut.read(second_location_add))
        self.assertEqual(data, await self.sut.read(item_location))
        await self._stop()
        with open(self._path + '/data00000.dat', 'rb') as f:
            x = f.read()
        self.assertEqual(x, self.sut._bake_new_file_header() + data)

        self._setup_sut()
        await self._run(expect_failure='Multiple checkpoint')
        await self.sut._write_db_checkpoint(checkpoint_id, 0)
        with self.assertRaises(exceptions.InvalidDBStateException):
            await self.sut._write_db_checkpoint(checkpoint_id, 0)
        self.assertTrue(self.sut.running)
        with self.assertRaises(exceptions.InvalidDBStateException):
            await self.sut._write_db_checkpoint(checkpoint_id + 1, 0)
        self.assertTrue(self.sut.running)
        shutil.copy(self._path + '/.checkpoint-333', self._path + '/.checkpoint-334')
        self.assertTrue(self.sut.running)
        with self.assertRaises(exceptions.NotRunningException) as e:
            for x in range(0, 20):
                await self.sut.add(os.urandom(1024 ** 2))
        self.assertFalse(self.sut.running)


class TestCleanStaleData(AioDiskDBTestCase):
    def test(self):
        with open(self._path + '/.tmp.data00000.dat', 'wb') as f:
            f.write(os.urandom(5))

        self._setup_sut()



from aiodiskdb import exceptions
from aiodiskdb.local_types import LockType


def ensure_running(expected_state: bool, allow_stop_state=False):
    def _decorator(f):
        async def _ensure(self, *a, **kw):
            if self._error:
                raise exceptions.NotRunningException('Database is in ERROR state')
            if expected_state and self._do_stop and not allow_stop_state:
                raise exceptions.NotRunningException('Database is in STOP state')
            if self.running != expected_state:
                raise exceptions.NotRunningException('Database it not running')
            return await f(self, *a, **kw)
        return _ensure
    return _decorator


def ensure_async_lock(lock_type: LockType):
    def _decorator(f):
        async def _ensure(self, *a, **kw):
            if lock_type == LockType.READ:
                await self._read_lock.acquire()
                self._incr_read()
                if self._reads_count == 1:
                    await self._write_lock.acquire()
                self._read_lock.release()
                try:
                    return await f(self, *a, **kw)
                finally:
                    await self._read_lock.acquire()
                    self._decr_read()
                    if not self._reads_count:
                        self._write_lock.release()
                    self._read_lock.release()
            elif lock_type == LockType.WRITE:
                in_transaction = self._transaction_lock.locked()
                if not in_transaction:
                    await self._transaction_lock.acquire()
                await self._write_lock.acquire()
                try:
                    return await f(self, *a, **kw)
                finally:
                    if not in_transaction:
                        self._transaction_lock.release()
                    self._write_lock.release()
            elif lock_type == LockType.TRANSACTION:
                await self._transaction_lock.acquire()
                try:
                    return await f(self, *a, **kw)
                finally:
                    self._transaction_lock.release()
            else:
                raise ValueError('Value must be LockType.READ or WRITE')
        return _ensure
    return _decorator


class GracefulExit(SystemExit):
    code = 1



import asyncio
import hashlib
import os
import shutil
from collections import OrderedDict
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
import time
import typing

from aiodiskdb import exceptions
from aiodiskdb.internals import ensure_running,  ensure_async_lock
from aiodiskdb.abstracts import AsyncRunnable, AioDiskDBTransactionAbstract
from aiodiskdb.local_types import ItemLocation, LockType, Buffer, TempBufferData, WriteEvent, FileHeader

_FILE_SIZE = 128
_FILE_PREFIX = 'data'
_FILE_ZEROS_PADDING = 5
_BUFFER_SIZE = 16
_BUFFER_ITEMS = 100
_FLUSH_INTERVAL = 30
_GENESIS_BYTES = b'\r\xce\x8f7'
_TIMEOUT = 30
_CONCURRENCY = 32


class AioDiskDB(AsyncRunnable):
    GENESIS_BYTES_LENGTH = 4
    HEADER_TRIM_OFFSET = 4
    RESERVED_HEADER_LENGTH = 16  # for future usage

    """
    Minimal on-disk DB, with buffers and timeouts.
    Made with love for Asyncio.
    """
    def __init__(
            self,
            path: str,
            create_if_not_exists: bool = False,
            overwrite: bool = False,
            clean_stale_data: bool = True,
            file_padding: int = _FILE_ZEROS_PADDING,
            file_prefix: str = _FILE_PREFIX,
            max_file_size: int = _FILE_SIZE,
            max_buffer_items: int = _BUFFER_ITEMS,
            max_buffer_size: int = _BUFFER_SIZE,
            flush_interval: int = _FLUSH_INTERVAL,
            genesis_bytes: bytes = _GENESIS_BYTES,
            timeout: int = _TIMEOUT,
            concurrency: int = _CONCURRENCY
    ):
        super().__init__()
        if not file_prefix.isalpha():
            raise exceptions.InvalidConfigurationException('Wrong file prefix (must be alphabetic string)')
        self.path = Path(path)
        if create_if_not_exists:
            self.path.mkdir(parents=True, exist_ok=True)
        self._file_prefix = file_prefix
        if max_file_size <= 0 or max_buffer_size <= 0:
            raise exceptions.InvalidConfigurationException('max file size and max buffer size must be > 0')
        if max_file_size < max_buffer_size:
            raise exceptions.InvalidConfigurationException('max_file_size must be >= max_buffer_size')
        self._file_padding = int(file_padding)
        self._max_file_size = int(float(max_file_size) * 1024 ** 2)
        self._max_buffer_items = int(max_buffer_items)
        self._max_buffer_size = int(float(max_buffer_size) * 1024 ** 2)
        self._flush_interval = int(flush_interval)
        if len(genesis_bytes) != self.GENESIS_BYTES_LENGTH:
            raise exceptions.InvalidConfigurationException('Genesis bytes length must be 4')
        self._genesis_bytes = genesis_bytes
        self._timeout = int(timeout)
        self._buffer_index = OrderedDict()
        self._buffers: typing.List[Buffer] = list()
        self._last_flush = None
        self._executor = ThreadPoolExecutor(max_workers=concurrency)
        self._overwrite = overwrite
        self._tmp_idx_and_buffer = TempBufferData(idx=dict(), buffer=None)
        self._file_header_size = self.GENESIS_BYTES_LENGTH + \
                             self.HEADER_TRIM_OFFSET + \
                             self.RESERVED_HEADER_LENGTH
        if clean_stale_data:
            self._drop_existing_temp_files()
            self._apply_checkpoint()
        else:
            self._ensure_no_pending_checkpoint()
        _max_accepted_file_size = 2**32 - 1 - self._file_header_size
        if max_file_size > _max_accepted_file_size:
            raise exceptions.InvalidConfigurationException(f'max file size is {_max_accepted_file_size}b')

    def _hash_file(self, f: typing.IO) -> typing.Optional[bytes]:
        """
        Hash files chunk by chunk, avoid to load the whole file in RAM.
        """
        i = 0
        chunk_size = 1024 ** 2
        _hash = None
        while 1:
            f.seek(i)
            c = f.read(chunk_size)
            if not c:
                break
            i += chunk_size
            _hash = hashlib.sha256().digest()
            f.seek(0)

        return _hash

    def _bake_new_file_header(self) -> bytes:
        """
        Bake a fresh file header for a new database file.
        """
        return FileHeader(
            genesis_bytes=self._genesis_bytes,
            trim_offset=0,
        ).serialize()

    def _read_file_header(self, f: typing.IO):
        """
        Read the first bytes of a file, and return the FileHeader
        """
        f.seek(0)
        header = f.read(self._file_header_size)
        if not self._is_file_header(header):
            raise exceptions.InvalidDataFileException
        return FileHeader(
            genesis_bytes=self._genesis_bytes,
            trim_offset=int.from_bytes(
                header[self.GENESIS_BYTES_LENGTH:self.GENESIS_BYTES_LENGTH+self.HEADER_TRIM_OFFSET],
                'little'
            )
        )

    def _bake_new_buffer(self, index: int):
        """
        Bake a fresh buffer, for a new database file.
        """
        return Buffer(
            index=index,
            data=b'',
            size=0,
            items=0,
            file_size=0,
            offset=0,
            head=True
        )

    async def _refresh_current_buffer(self):
        """
        Reload the current session buffer from the system state.
        To be used after a transaction or an index change (trim\drop).
        """
        self._buffer_index.pop(self._buffers[-1].index)
        self._buffers = []
        await self._setup_current_buffer()

    def _pre_stop_signal(self) -> bool:
        """
        Handle graceful stop signals. Flush buffer to disk.
        """
        if self._blocking_stop:
            return False
        self._blocking_stop = True
        assert not self._tmp_idx_and_buffer.idx
        self._tmp_idx_and_buffer = None
        while self._buffer_index:
            buffer = self._buffers.pop(0)
            v = self._buffer_index.pop(buffer.index)
            self._save_buffer_to_disk(TempBufferData(idx={buffer.index: v}, buffer=buffer))
        return True

    @ensure_async_lock(LockType.WRITE)
    async def _clean_temp_buffer(self):
        """
        Cleanup the current temp buffer. Unload the RAM, to be triggered after a disk flush.
        """
        self._tmp_idx_and_buffer = TempBufferData(idx=dict(), buffer=None)

    def _read_data_from_buffer(self, location: ItemLocation):
        """
        Read data from the main buffer.
        Data is here until the flush_buffer task triggers a disk flush.
        """
        try:
            idx = self._buffer_index[location.index][location.position]
            buffer = self._buffers[idx]
        except KeyError:
            return None

        buffer_shift = 0 if buffer.size == buffer.file_size else buffer.file_size - buffer.size
        relative_position = location.position - buffer_shift
        data = buffer.data[relative_position: relative_position + location.size]
        assert len(data) == location.size, f'{len(data)} != {location.size}'
        return data

    def _read_data_from_temp_buffer(self, location: ItemLocation):
        """
        Data is placed into the temp buffer while saving it on disk,
        so that writes are non blocking.
        """
        try:
            idx = self._tmp_idx_and_buffer.idx[location.index][location.position]
        except KeyError:
            return None

        buffer = self._tmp_idx_and_buffer.buffer
        # just ensures the idx was previously saved, the temp buffer is flat.
        buffer_shift = 0 if buffer.size == buffer.file_size else buffer.file_size - buffer.size
        relative_position = location.position - buffer_shift
        data = idx is not None and buffer.data[relative_position:relative_position + location.size] or None
        assert data is not None and len(data) == location.size or data is None
        return data

    def _get_filename_by_idx(self, idx: int, temp=False) -> str:
        t = temp and '.tmp.' or ''
        return f'{self.path}/{t}{self._file_prefix}' + f'{idx}'.zfill(self._file_padding) + '.dat'

    async def _setup_current_buffer(self):
        """
        Setup the current buffer, starting from the disk files.
        If no files are found, setup a fresh buffer, otherwise check the genesis bytes.
        """
        files = sorted(os.listdir(self.path))
        last_file = files and list(
            filter(lambda x: x.startswith(self._file_prefix) and x.endswith('.dat'), files)
        )[-1]
        if not last_file:
            # No previous files found for the current setup. Starting a new database.
            buffer = self._bake_new_buffer(0)
            offset, curr_idx = 0, 0
        else:
            curr_idx = int(last_file.replace(self._file_prefix, '').replace('.dat', ''))
            filename = self._get_filename_by_idx(curr_idx)
            curr_size = os.path.getsize(filename) - self._file_header_size
            if curr_size < 0:
                raise exceptions.InvalidDataFileException('Invalid file size')
            with open(filename, 'rb') as f:
                header = self._read_file_header(f)
                offset = header.trim_offset
            data = b''
            buffer = Buffer(
                index=curr_idx, data=data, size=len(data),
                items=0, file_size=curr_size, offset=offset,
                head=False
            )
        self._buffers.append(buffer)
        self._buffer_index[curr_idx] = OrderedDict()

    @ensure_async_lock(LockType.TRANSACTION)
    async def _teardown(self):
        self._buffers[-1].size and await self._flush_buffer_no_transaction_lock()
        self._executor.shutdown(wait=True)

    def enable_overwrite(self):
        self._overwrite = True

    def disable_overwrite(self):
        self._overwrite = False

    @ensure_async_lock(LockType.WRITE)
    async def _pop_buffer_data(self) -> TempBufferData:
        """
        Remove the buffer from the data queue.
        Put it into the temp storage for disk writing.
        Allocate a new buffer for the data queue.
        """
        assert not self._tmp_idx_and_buffer.buffer, 'wrong state, cannot recover. buffer lost.'
        buffer = self._buffers.pop(0)
        v = self._buffer_index.pop(buffer.index)
        self._tmp_idx_and_buffer = TempBufferData(idx={buffer.index: v}, buffer=buffer)
        if buffer.file_size > self._max_file_size:
            new_buffer = self._bake_new_buffer(buffer.index + 1)
        else:
            new_buffer = Buffer(
                index=buffer.index, data=b'',
                size=0, items=0, file_size=buffer.file_size,
                offset=buffer.offset, head=False
            )
        self._buffers.append(new_buffer)
        self._buffer_index[new_buffer.index] = OrderedDict()
        return self._tmp_idx_and_buffer

    def _save_buffer_to_disk(self, buffer_data: TempBufferData):
        """
        Actually saves data from a temp buffer to the target file and position.
        """
        assert buffer_data.buffer and buffer_data.idx, (buffer_data.buffer, buffer_data.idx)
        buffer = buffer_data.buffer
        filename = self._get_filename_by_idx(buffer.index)
        try:
            if buffer.head:
                if Path(filename).exists():
                    self._stop = True
                    raise exceptions.FilesInconsistencyException(f'File {filename} should not exists.')

            elif os.path.getsize(filename) != buffer.file_size - buffer.size + self._file_header_size:
                self._stop = True
                raise exceptions.InvalidDataFileException(f'File {filename} has unexpected size.')

            with open(filename, 'ab') as file:
                if buffer.head:
                    file.write(self._bake_new_file_header())
                file.write(buffer.data)
        except FileNotFoundError:
            self._stop = True
            raise exceptions.FilesInconsistencyException(f'Missing file {filename}')

    def _is_file_header(self, data: bytes) -> bool:
        """
        Verify the header correctness
        """
        if len(data) != self._file_header_size:
            return False
        p2 = self.GENESIS_BYTES_LENGTH + self.HEADER_TRIM_OFFSET
        if not data[:self.GENESIS_BYTES_LENGTH] == self._genesis_bytes:
            return False
        if not int.from_bytes(
                data[self.GENESIS_BYTES_LENGTH:p2], 'little'
        ) < self._max_file_size - self._file_header_size:
            return False
        if not data[p2:p2+self.RESERVED_HEADER_LENGTH] == int(0).to_bytes(self.RESERVED_HEADER_LENGTH, 'little'):
            return False
        return True

    @ensure_async_lock(LockType.TRANSACTION)
    async def _flush_buffer(self):
        """
        Trigger blocking\non-blocking operations.
        - pop_buffer: coroutine, locked for writing
        - save_buffer_to_disk: blocking threaded task, non locked
        - clean_temp_buffer: coroutine, locked for writing
        """
        await self._flush_buffer_no_transaction_lock()

    async def _flush_buffer_no_transaction_lock(self):
        timestamp = int(time.time()*1000)
        temp_buffer_data = await self._pop_buffer_data()
        await self._write_db_checkpoint(timestamp, temp_buffer_data.buffer.index)
        await asyncio.get_event_loop().run_in_executor(
            self._executor,
            self._save_buffer_to_disk,
            temp_buffer_data
        )
        flush_time = time.time()
        if self.events.on_write:
            offset = temp_buffer_data.buffer.offset
            asyncio.get_event_loop().create_task(
                self.events.on_write(
                    flush_time,
                    WriteEvent(
                        index=temp_buffer_data.buffer.index,
                        position=temp_buffer_data.buffer.file_size - temp_buffer_data.buffer.size + offset,
                        size=temp_buffer_data.buffer.size
                    )
                )
            )

        await self._clean_temp_buffer()
        self._last_flush = flush_time
        await self._clean_db_checkpoint(timestamp)

    def _process_location_read(self, location: ItemLocation):
        """
        Actually read data from files.
        """
        try:
            filename = self._get_filename_by_idx(location.index)
            with open(filename, 'rb') as file:
                header = self._read_file_header(file)
                if location.position < header.trim_offset:
                    return None
                file.seek(self._file_header_size + location.position - header.trim_offset)
                read = file.read(location.size)
            return read or None
        except FileNotFoundError:
            return None

    async def _pre_loop(self):
        if self._last_flush is None:
            self._last_flush = time.time()
        await self._setup_current_buffer()

    async def _run_loop(self):
        buffer = self._buffers[0]
        if buffer.file_size >= self._max_file_size:
            await self._flush_buffer()
        elif buffer.size >= self._max_buffer_size:
            await self._flush_buffer()
        elif time.time() - self._last_flush >= self._flush_interval:
            await self._flush_buffer()
        elif buffer.items >= self._max_buffer_items:
            await self._flush_buffer()

    @ensure_running(True)
    async def add(self, data: bytes) -> ItemLocation:
        """
        Enqueue data into a buffer, update the indexes for reads from RAM.

        :param data: bytes
        :return: ItemLocation(int, int, int)
        """
        if not data:
            raise exceptions.EmptyPayloadException
        s = time.time()
        while self._buffers[-1].file_size + self._buffers[-1].offset >= self._max_file_size \
                or self._buffers[-1].size >= self._max_buffer_size:
            # wait for the LRT to shift the buffers
            await asyncio.sleep(0.01)
            if time.time() - s > self._timeout:
                raise exceptions.TimeoutException
        return await self._add(data)

    @ensure_async_lock(LockType.WRITE)
    async def _add(self, data: bytes) -> ItemLocation:
        """
        Add data into the current buffer.
        """
        buffer = self._buffers[-1]
        data_size = len(data)
        if data_size > self._max_buffer_size:
            raise exceptions.WriteFailedException(
                f'File too big: {data_size} > {self._max_buffer_size}'
            )
        location = ItemLocation(
            index=buffer.index,
            position=buffer.file_size,
            size=data_size
        )
        self._buffer_index[location.index][location.position] = len(self._buffers) - 1
        buffer.data += data
        buffer.items += 1
        buffer.size += data_size
        buffer.file_size += data_size
        return location

    @ensure_running(True)
    @ensure_async_lock(LockType.READ)
    async def read(self, location: ItemLocation):
        """
        Reads data from the DB.
        If there's no data in RAM for the given location, try with a disk read.

        :param location: ItemLocation(int, int, int)
        :return: bytes
        """
        return self._read_data_from_buffer(location) or \
            self._read_data_from_temp_buffer(location) or \
            await asyncio.get_event_loop().run_in_executor(
            self._executor,
            self._process_location_read,
            location
        )

    @ensure_running(True)
    async def transaction(self) -> AioDiskDBTransactionAbstract:
        from aiodiskdb.transaction import AioDiskDBTransaction
        return AioDiskDBTransaction(self)

    def destroy_db(self):
        """
        Destroy the DB, clean the disk.
        """
        if self.running:
            raise exceptions.RunningException('Database must be stopped before destroying it')
        if not self._overwrite:
            raise exceptions.ReadOnlyDatabaseException
        shutil.rmtree(self.path)
        return True

    @ensure_running(True)
    @ensure_async_lock(LockType.TRANSACTION)
    async def drop_index(self, index: int) -> int:
        """
        Drop a single index.
        Ensures a flush first.
        If the deleted index is the current one, setup it again from scratch.
        """
        if not self._overwrite:
            raise exceptions.ReadOnlyDatabaseException
        assert not self._tmp_idx_and_buffer.buffer
        return await self._drop_index_non_locked(index)

    async def _drop_index_non_locked(self, index: int):
        filename = self._get_filename_by_idx(index)
        if not Path(filename).exists():
            raise exceptions.IndexDoesNotExist
        dropped_index_size = os.path.getsize(filename) - self._file_header_size
        os.remove(filename)
        if self._buffers[-1].index == index:
            await self._refresh_current_buffer()

        if self.events.on_index_drop:
            asyncio.get_event_loop().create_task(
                self.events.on_index_drop(time.time(), index, dropped_index_size)
            )
        return dropped_index_size

    @ensure_running(True, allow_stop_state=True)
    async def _write_db_checkpoint(self, timestamp: int, *files_indexes: int):
        """
        Persist the current status of files before appending data.
        It will be reverted in case of a commit failure.
        """
        checkpoints = self._get_checkpoints()
        if checkpoints:
            raise exceptions.InvalidDBStateException('Requested a checkpoint, but a checkpoint already exist')

        filenames = map(self._get_filename_by_idx, files_indexes)

        checkpoint = []
        for i, file_name in enumerate(filenames):
            try:
                file_size = os.path.getsize(file_name)
            except FileNotFoundError:
                continue
            with open(file_name, 'rb') as f:
                file_hash = self._hash_file(f)
                f.seek(max(0, file_size - 8))
                last_file_bytes = f.read(8)
            checkpoint.append(
                [
                    files_indexes[i].to_bytes(6, 'little'),
                    file_size.to_bytes(6, 'little'),
                    last_file_bytes,
                    file_hash
                ]
            )
        with open(os.path.join(self.path, f'.checkpoint-{timestamp}'), 'wb') as f:
            for s in checkpoint:
                f.write(b''.join(s))

    @ensure_running(True, allow_stop_state=True)
    async def _clean_db_checkpoint(self, timestamp: int):
        """
        The transaction is successfully committed.
        The checkpoint file must be deleted.
        """
        try:
            os.remove(os.path.join(self.path, f'.checkpoint-{timestamp}'))
        except FileNotFoundError as e:
            raise exceptions.InvalidDBStateException(
                'Requested to delete a checkpoint that does not exist'
            ) from e

    def _get_checkpoints(self) -> typing.List[str]:
        checkpoints = list(
            filter(
                lambda x: x.startswith('.checkpoint-'),
                os.listdir(self.path)
            )
        )
        if len(checkpoints) > 1:
            # This should NEVER happen.
            raise exceptions.InvalidDBStateException('Multiple checkpoints, unrecoverable error.')
        return checkpoints

    def _apply_checkpoint(self):
        """
        Apply a database checkpoint to recover a previous state.
        A checkpoint is created before a transaction commit is made.
        A checkpoint is created before a transaction commit is made.
        Having a checkpoint file during the AioDiskDB boot mean that a transaction commit is failed,
        and stale data may exists in the files.
        Applying the checkpoint discard all the data added to the files after taking it.
        Stale data is deleted.
        """
        assert not self.running
        checkpoints = self._get_checkpoints()
        if not checkpoints:
            return
        checkpoint_file = checkpoints[0]
        checkpoint_filename = os.path.join(str(self.path), checkpoint_file)
        with open(checkpoint_filename, 'rb') as f:
            checkpoint = f.read()
        assert checkpoint
        pos = 0
        files = []
        while pos < len(checkpoint):
            files.append(
                [
                    int.from_bytes(checkpoint[pos:pos+6], 'little'),
                    int.from_bytes(checkpoint[pos+6:pos+12], 'little'),
                    checkpoint[pos+12:pos+20],
                    checkpoint[pos+20:pos+52]
                ]
            )
            pos += 52
        for file_data in files:
            self._recover_files_from_checkpoint(file_data)
        os.remove(checkpoint_filename)

    def _recover_files_from_checkpoint(self, file_data: typing.List):
        """
        Actually apply checkpoint rules to existing files.
        """
        file_id = file_data[0]
        checkpoint_file_size = file_data[1]
        checkpoint_file_last_bytes = file_data[2]
        expected_hash = file_data[3]
        origin_file_name = self._get_filename_by_idx(file_id)
        bkp_file_name = self._get_filename_by_idx(file_id, temp=True)
        shutil.copy(origin_file_name, bkp_file_name)
        exc = None
        with open(bkp_file_name, 'r+b') as f:
            header = f.read(self._file_header_size)
            if not self._is_file_header(header):
                exc = exceptions.InvalidDataFileException('Invalid file header')
            else:
                f.seek(checkpoint_file_size - len(checkpoint_file_last_bytes))
                data = f.read(len(checkpoint_file_last_bytes))
                f.seek(0)
                f.truncate(checkpoint_file_size)
                if data != checkpoint_file_last_bytes:
                    exc = exceptions.InvalidDataFileException('File corrupted. Unrecoverable error')
                final_hash = self._hash_file(f)
                if final_hash != expected_hash:
                    exc = exceptions.InvalidDataFileException('Invalid file hash recovered')
        if not exc:
            shutil.copy(bkp_file_name, origin_file_name)
        os.remove(bkp_file_name)
        if exc:
            raise exc

    def _drop_existing_temp_files(self):
        for file in filter(
            lambda x: x.startswith(f'.tmp.{self._file_prefix}') and x.endswith('.dat'),
            os.listdir(self.path)
        ):
            os.remove(os.path.join(str(self.path), file))

    def _ensure_no_pending_checkpoint(self):
        assert not self.running
        if any(map(
            lambda x: x.startswith('.checkpoint-'),
            os.listdir(self.path)
        )):
            raise exceptions.InvalidDBStateException('Pending checkpoint. DB must be cleaned.')

    @ensure_running(True)
    @ensure_async_lock(LockType.TRANSACTION)
    async def rtrim(self, index: int, trim_from: int, safety_check: bytes = b'') -> int:
        """
        Trim an index, from the right.

        trim_from: the index location from which to delete data.
        safety_check: optional, must match the first bytes of the trimmed slice.

        return the size of the trimmed slice.
        """
        if not self._overwrite:
            raise exceptions.ReadOnlyDatabaseException
        assert not self._tmp_idx_and_buffer.buffer, self._tmp_idx_and_buffer.buffer
        try:
            await self._flush_buffer_no_transaction_lock()
            return await self._do_rtrim(index, trim_from, safety_check)
        except FileNotFoundError:
            raise exceptions.IndexDoesNotExist

    async def _do_rtrim(self, index: int, trim_from: int, safety_check: bytes) -> int:
        if index < 0:
            raise exceptions.IndexDoesNotExist('Index must be > 0')
        filename = self._get_filename_by_idx(index)
        assert isinstance(trim_from, int)
        if not trim_from and not safety_check:
            return await self._drop_index_non_locked(index)

        elif not trim_from:
            with open(filename, 'r+b') as f:
                if safety_check != f.read(len(safety_check)):
                    raise exceptions.InvalidTrimCommandException('safety check failed')
            return await self._drop_index_non_locked(index)

        pre_trim_file_size = os.path.getsize(filename)
        with open(filename, 'r+b') as f:
            f.seek(trim_from + self._file_header_size)
            if safety_check:
                if safety_check != f.read(len(safety_check)):
                    raise exceptions.InvalidTrimCommandException('safety check failed')
            f.seek(trim_from + self._file_header_size)
            f.truncate()
        file_size = os.path.getsize(filename)
        if self._buffers[-1].index == index:
            await self._refresh_current_buffer()
        return pre_trim_file_size - file_size

    @ensure_running(True)
    @ensure_async_lock(LockType.TRANSACTION)
    async def ltrim(self, index: int, trim_to: int, safety_check: bytes = b''):
        """
        Trim an index, from the left.

        trim_to: the index location of data that are going to be kept,
        anything before this point is trimmed out.
        safety_check: optional, must match the last bytes of the trimmed slice.
        """
        if not self._overwrite:
            raise exceptions.ReadOnlyDatabaseException

        if index < 0:
            raise exceptions.InvalidDataFileException(
                'Index must be >= 0'
            )
        assert not self._tmp_idx_and_buffer.buffer, self._tmp_idx_and_buffer.buffer
        filename = self._get_filename_by_idx(index)
        temp_filename = self._get_filename_by_idx(index, temp=True)
        try:
            os.path.getsize(temp_filename)
            exc = exceptions.InvalidDBStateException('trim file already exists')
            self._error = exc
            self._stop = True
        except FileNotFoundError:
            pass
        try:
            await self._flush_buffer_no_transaction_lock()
            with open(filename, 'rb') as origin:
                header = self._read_file_header(origin)
            if trim_to <= header.trim_offset:
                raise exceptions.InvalidTrimCommandException(
                    f'trim_to must be > current_offset ({header.trim_offset})'
                )

            await self._do_ltrim(index, trim_to, safety_check, header)
            if self._buffers[-1].index == index:
                await self._refresh_current_buffer()
        except FileNotFoundError:
            raise exceptions.IndexDoesNotExist

    async def _do_ltrim(self, index: int, trim_to: int, safety_check: bytes, header: FileHeader) -> int:
        filename = self._get_filename_by_idx(index)
        index_size = os.path.getsize(filename) - self._file_header_size + header.trim_offset
        if index_size < trim_to:
            raise exceptions.InvalidTrimCommandException('trim_to must be <= index_size')
        elif index_size == trim_to:
            return await self._drop_index_non_locked(index)

        with open(filename, 'rb') as origin:
            safety_check_length = len(safety_check)
            seek_at = self._file_header_size + trim_to - header.trim_offset
            if safety_check_length:
                origin.seek(seek_at - safety_check_length)
                check = origin.read(safety_check_length)
                if check != safety_check:
                    raise exceptions.InvalidTrimCommandException('safety check failed')
            else:
                origin.seek(seek_at)
            temp_filename = self._get_filename_by_idx(index, temp=True)
            with open(temp_filename, 'wb') as target:
                header.trim_offset = trim_to
                target.write(header.serialize())
                data = origin.read(1024**2)
                while data:
                    target.write(data)
                    data = origin.read(1024 ** 2)
        shutil.move(temp_filename, filename)
        return trim_to



import asyncio
from dataclasses import dataclass

import typing
from enum import Enum


@dataclass
class ItemLocation:
    index: int
    position: int
    size: int

    def deserialize(self):
        return self.index.to_bytes(4, 'little') + \
            self.position.to_bytes(4, 'little') + \
            self.size.to_bytes(4, 'little')

    @classmethod
    def serialize(cls, location: bytes):
        return cls(
            index=int.from_bytes(location[:4], 'little'),
            position=int.from_bytes(location[4:8], 'little'),
            size=int.from_bytes(location[8:12], 'little'),
        )


class LockType(Enum):
    READ = 0
    WRITE = 1
    TRANSACTION = 2


@dataclass
class Buffer:
    index: int
    data: bytes
    size: int
    items: int
    file_size: int
    offset: int
    head: bool


@dataclass
class TempBufferData:
    buffer: typing.Optional[Buffer]
    idx: typing.Dict


@dataclass
class EventsHandlers:
    """
    Callback signature, first argument is always the execution timestamp (time.time()).

    async def callback(fired_at: int, *callback_data):
        pass
    """
    on_start: typing.Optional[callable] = None
    on_stop: typing.Optional[callable] = None
    on_failure: typing.Optional[callable] = None
    on_index_drop: typing.Optional[callable] = None
    on_write: typing.Optional[callable] = None

    def __setattr__(self, key, value):
        if value is not None and not asyncio.iscoroutinefunction(value):
            raise TypeError(f'{key} must be a coroutine')
        self.__dict__[key] = value


class TransactionStatus(Enum):
    INITIALIZED = 1
    ONGOING = 2
    DONE = 3


@dataclass
class WriteEvent:
    index: int
    position: int
    size: int


@dataclass
class FileHeader:
    genesis_bytes: bytes
    trim_offset: int

    def serialize(self) -> bytes:
        return self.genesis_bytes + \
               int(self.trim_offset).to_bytes(4, 'little') + \
               int(0).to_bytes(16, 'little')  # reserved



import asyncio
from aiodiskdb.local_types import ItemLocation, WriteEvent
from test import AioDiskDBTestCase, run_test_db


class TestReadWriteCached(AioDiskDBTestCase):
    @run_test_db
    async def test(self):
        item_location = await self.sut.add(b'test_1')
        self.assertEqual(
            ItemLocation(0, 0, 6),
            item_location
        )
        item_location_2 = await self.sut.add(b'test_2')
        self.assertEqual(
            ItemLocation(0, 6, 6),
            item_location_2
        )
        read1 = await self.sut.read(item_location)
        self.assertEqual(b'test_1', read1)
        read2 = await self.sut.read(item_location_2)
        self.assertEqual(b'test_2', read2)
        self.assertEqual(self._writes, [])

    def tearDown(self):
        self.assertEqual(1, len(self._stops))
        self.assertIsInstance(self._stops[0][0], float)
        self.assertEqual(1, len(self._starts))
        self.assertIsInstance(self._starts[0][0], float)
        self.assertEqual(1, len(self._writes))
        self.assertIsInstance(self._writes[0][0], float)
        self.assertEqual(
            WriteEvent(index=0, position=0, size=12),
            self._writes[0][1]
        )
        super().tearDown()


class TestReadWriteNonCached(AioDiskDBTestCase):
    @run_test_db
    async def test(self):
        item_location = await self.sut.add(b'test_1')
        self.assertEqual(
            ItemLocation(0, 0, 6),
            item_location
        )
        item_location_2 = await self.sut.add(b'test_2')
        self.assertEqual(
            ItemLocation(0, 6, 6),
            item_location_2
        )
        await self.sut.stop()  # stop the sut, ensures the data write
        self.assertEqual(
            WriteEvent(index=0, position=0, size=12),
            self._writes[0][1]
        )
        self._setup_sut()  # re-instance the sut from scratch.
        self.loop.create_task(self.sut.run())
        while not self.sut.running:
            await asyncio.sleep(0.01)
        new_loc_2 = ItemLocation.serialize(item_location_2.deserialize())
        read1 = await self.sut.read(new_loc_2)
        self.assertEqual(b'test_2', read1)
        read2 = await self.sut.read(item_location_2)
        self.assertEqual(b'test_2', read2)

        item_location_3 = item_location_2
        item_location_3.index = 99
        self.assertEqual(None, await self.sut.read(item_location_3))


class TestFlushInterval(AioDiskDBTestCase):
    def setUp(self, *a):
        super().setUp(*a, flush_interval=3)

    @run_test_db
    async def test(self):
        await self.sut.add(b'test_2')
        self.assertEqual(6, self.sut._buffers[-1].size)
        await asyncio.sleep(self.sut._flush_interval + 1)
        self.assertEqual(0, self.sut._buffers[-1].size)



import asyncio
import os
import time
from random import randint

from test import run_test_db, AioDiskDBConcurrencyTest


class TestAioDiskDBConcurrentReadWrite(AioDiskDBConcurrencyTest):

    @run_test_db
    async def test(self):
        self._running_test = True
        self.loop.create_task(self._random_reads())
        data_stored = dict()
        total_size = 0
        s = time.time()
        sizes = []
        while sum(data_stored.values()) < 1000 * 1024 ** 2:
            size = randint(1024, 1024**2)
            sizes.append(size)
            data = os.urandom(size)
            location = await self.sut.add(data)
            self._writes_count += 1
            data_stored.setdefault(location.index, 0)
            self._data.append([location, data])
            data_stored[location.index] += size
            total_size += size
            await asyncio.sleep(0.00001)
            self.assertTrue(self._ongoing_reads, msg='reads failed')
        self._pause_reads = True
        print(f'R/W concurrency test over. Duration: {time.time() - s:.2f}s, '
              f'Reads: {self._reads_count}, Writes: {self._writes_count}, '
              f'Bandwidth: {total_size // 1024 ** 2}MB, '
              f'Avg file size: {sum(sizes) / len(sizes) // 1024}kB'
              )

        current_reads = self._reads_count
        # test is over, repeat the random reads with a new DB instance on the same data.
        print('Read only test with no-cache instance:')
        while self._ongoing_reads:
            await asyncio.sleep(1)
        await self.sut.stop()
        self._setup_sut()
        self.loop.create_task(self.sut.run())
        while not self.sut.running:
            await asyncio.sleep(0.01)
        self._pause_reads = False
        s = time.time()
        while time.time() - s < 10:
            await asyncio.sleep(2)
            self.assertTrue(self._ongoing_reads, msg='reads failed')
        self._stop_reads = True
        print(f'Read only test from disk over. Reads: {self._reads_count - current_reads}')




import asyncio
import time

import typing
from collections import OrderedDict, deque

from aiodiskdb import AioDiskDB, exceptions
from aiodiskdb.abstracts import AioDiskDBTransactionAbstract
from aiodiskdb.internals import ensure_async_lock
from aiodiskdb.local_types import TempBufferData, TransactionStatus, Buffer, ItemLocation, LockType


class AioDiskDBTransaction(AioDiskDBTransactionAbstract):
    def __init__(self, session: AioDiskDB):
        self.session = session
        self._stack = deque()
        self._status = TransactionStatus.INITIALIZED
        self._lock = asyncio.Lock()
        self._locations = list()

    @property
    def _transaction_lock(self):
        return self.session._transaction_lock

    def _bake_new_temp_buffer_data(self, res: typing.List[TempBufferData]) -> None:
        """
        The current buffer must be full if this method is called.
        Bake a new buffer contiguous to the current.
        """
        new_idx = res[-1].buffer.index + 1
        res.append(
            TempBufferData(
                self.session._bake_new_buffer(new_idx),
                idx={new_idx: dict()}
            )
        )

    def _bake_temp_buffer_data(self) -> typing.List[TempBufferData]:
        """
        Bake buffer data so that <session._save_data_to_disk>
        function could handle it with no changes.
        There is no buffer size limit while in a Transaction,
        only the file size limit is respected.
        """
        assert self.session._buffers[-1].size == 0
        current_buffer = self.session._buffers[-1]
        res = [
            TempBufferData(
                buffer=current_buffer,
                idx={current_buffer.index: self.session._buffer_index}
            )
        ]
        max_file_size = self.session._max_file_size
        while len(self._stack):
            data = self._stack.popleft()
            data_size = len(data)
            buffer: Buffer = res[-1].buffer
            if data_size + buffer.file_size > max_file_size:
                self._bake_new_temp_buffer_data(res)
                self._stack.insert(0, data)
                continue
            self._locations.append(
                ItemLocation(buffer.index, buffer.file_size, data_size)
            )
            buffer.data += data
            buffer.size += data_size
            buffer.file_size += data_size
        return res

    async def _update_session_buffer(self, temp_buffer_data: TempBufferData):
        """
        This is fired after a Transaction is successfully saved to disk.
        Set the session buffer to the latest baked by the Transaction.
        """
        temp_buffer_data.buffer.data = b''
        temp_buffer_data.buffer.size = 0
        temp_buffer_data.buffer.head = not temp_buffer_data.buffer.file_size
        self.session._buffers[-1] = temp_buffer_data.buffer
        self.session._buffer_index = OrderedDict({temp_buffer_data.buffer.index: dict()})

    def add(self, data: bytes):
        """
        Add some data to a transaction.
        Data added into this scope is not available into the session
        until the transaction is committed.
        """
        if len(data) > self.session._max_file_size:
            raise exceptions.WriteFailedException(
                f'File too big: {len(data)} > {self.session._max_file_size}'
            )
        if self._status == TransactionStatus.DONE:
            raise exceptions.TransactionAlreadyCommittedException
        elif self._status == TransactionStatus.ONGOING:
            raise exceptions.TransactionCommitOnGoingException

        self._stack.append(data)

    async def commit(self) -> typing.Iterable[ItemLocation]:
        """
        Commit a transaction, save to data the <_stack> content, using TempBufferData objects.
        """
        await self._lock.acquire()
        try:
            if self._status == TransactionStatus.DONE:
                raise exceptions.TransactionAlreadyCommittedException
            elif not self._stack:
                raise exceptions.EmptyTransactionException
            now = int(time.time()*1000)

            await self.session._flush_buffer()
            await self._do_commit(now)
            locations = self._locations
            self._locations = list()
            return locations
        finally:
            self._lock.release()

    @ensure_async_lock(LockType.TRANSACTION)
    async def _do_commit(self, timestamp: int):
        """
        Part of the <commit> method.
        Actually saves data to disk.
        """
        self._status = TransactionStatus.ONGOING
        assert not self.session._buffers[-1].size
        idx_involved_in_batch = list()
        temp_buffers_data = self._bake_temp_buffer_data()
        for buff in temp_buffers_data:
            idx_involved_in_batch.extend(list(buff.idx))
        await self.session._write_db_checkpoint(
            timestamp, *set(idx_involved_in_batch)
        )
        assert temp_buffers_data
        temp_buffer_data = None
        for temp_buffer_data in temp_buffers_data:
            await asyncio.get_event_loop().run_in_executor(
                None,
                self.session._save_buffer_to_disk,
                temp_buffer_data
            )
        assert temp_buffer_data
        await self._update_session_buffer(temp_buffer_data)
        await self.session._clean_db_checkpoint(timestamp)
        self._status = TransactionStatus.DONE



from aiodiskdb import exceptions
from test import AioDiskDBTestCase, run_test_db


class AioDBTestDropIndex(AioDiskDBTestCase):
    def setUp(self, *a, **kw):
        super().setUp(max_file_size=1, max_buffer_size=1, overwrite=False)

    async def test(self):
        await self._run()
        with self.assertRaises(exceptions.ReadOnlyDatabaseException):
            await self.sut.drop_index(99)
        with self.assertRaises(exceptions.RunningException):
            self.sut.destroy_db()
        await self.sut.stop()
        with self.assertRaises(exceptions.ReadOnlyDatabaseException):
            self.sut.destroy_db()
        self._setup_sut()
        await self._run()
        self.sut.enable_overwrite()
        with self.assertRaises(exceptions.IndexDoesNotExist):
            await self.sut.drop_index(99)
        transaction = await self.sut.transaction()
        transaction.add(b'cafe')
        t_loc = await transaction.commit()
        self.assertTrue(bool(self.sut._buffers))
        loc_0 = await self.sut.add(b'babe')
        self.assertEqual(b'babe', await self.sut.read(loc_0))
        self.assertEqual(b'cafe', await self.sut.read(t_loc[0]))
        await self.sut.drop_index(0)
        self.assertIsNone(await self.sut.read(loc_0))
        self.assertIsNone(await self.sut.read(t_loc[0]))
        loc = await self.sut.add(b'test_after')
        d = await self.sut.read(loc)
        self.assertEqual(d, b'test_after')
        await self._stop()
        self._setup_sut()
        self.sut.enable_overwrite()
        await self._run()
        await self.sut.drop_index(0)
        await self._stop()

    def tearDown(self) -> None:
        self.assertEqual(2, len(self._index_drops))
        self.assertIsInstance(self._index_drops[0][0], float)
        self.assertEqual(self._index_drops[0][1], 0)  # index
        self.assertEqual(self._index_drops[0][2], 4)  # length of 'cafe'
        super().tearDown()



import os
import time
from pathlib import Path

from aiodiskdb import exceptions
from test import AioDiskDBTestCase


class AioDBTestErrorWrongFiles(AioDiskDBTestCase):
    def setUp(self, *a, **kw):
        super().setUp(max_file_size=1, max_buffer_size=1)

    async def test(self):
        await self._run()
        with self.assertRaises(exceptions.EmptyPayloadException):
            await self.sut.add(b'')
        b = os.urandom(1024 ** 2 + 10)
        with self.assertRaises(exceptions.WriteFailedException):
            await self.sut.add(b)
        await self._stop()

    def tearDown(self) -> None:
        self.assertEqual(1, len(self._stops))
        self.assertIsInstance(self._stops[0][0], float)
        self.assertEqual(1, len(self._starts))
        self.assertIsInstance(self._starts[0][0], float)
        self.assertEqual(0, len(self._failures))
        self.assertEqual(0, len(self._writes))
        super().tearDown()


class AioDBTestErrorWrongHeaderShort(AioDiskDBTestCase):
    def setUp(self, *a, **kw):
        super().setUp()
        Path(self._path).mkdir(parents=True, exist_ok=True)
        with open(self._path + '/data00000.dat', 'wb') as f:
            f.write(b'aa'*8)

    async def test(self):
        with self.assertRaises(exceptions.InvalidDataFileException):
            await self.sut.run()


class AioDBTestErrorWrongHeader(AioDiskDBTestCase):
    def setUp(self, *a, **kw):
        super().setUp()
        Path(self._path).mkdir(parents=True, exist_ok=True)
        with open(self._path + '/data00000.dat', 'wb') as f:
            f.write(b'aa'*16)

    async def test(self):
        with self.assertRaises(exceptions.InvalidDataFileException):
            await self.sut.run()


class AioDBTestErrorWrongGenesisFileShouldNotExists(AioDiskDBTestCase):
    def setUp(self, *a, **kw):
        super().setUp(max_file_size=0.1, max_buffer_size=0.1)

    def _corrupt_file(self):
        Path(self._path).mkdir(parents=True, exist_ok=True)
        with open(self._path + '/data00001.dat', 'wb') as f:
            f.write(b'aa'*8)

    async def test(self):
        await self._run(expect_failure='data00001.dat should not exists')
        self._corrupt_file()
        with self.assertRaises(exceptions.NotRunningException):
            for _ in range(0, 100):
                await self.sut.add(os.urandom(10240))
        with self.assertRaises(exceptions.NotRunningException) as e:
            await self.sut.run()
        self.assertTrue('ERROR state' in str(e.exception))

    def tearDown(self):
        self.assertEqual(1, len(self._stops))
        self.assertIsInstance(self._stops[0][0], float)
        self.assertEqual(1, len(self._starts))
        self.assertIsInstance(self._starts[0][0], float)
        self.assertEqual(1, len(self._failures))
        self.assertIsInstance(self._failures[0][0], float)
        self.assertTrue(self._failures[0][0] - time.time() < 2)
        super().tearDown()


class AioDBTestErrorZeroDBSizeError(AioDiskDBTestCase):
    async def test(self):
        with self.assertRaises(exceptions.InvalidConfigurationException):
            super().setUp(max_file_size=0, max_buffer_size=0)


class AioDBTestErrorInvalidDBSizeError(AioDiskDBTestCase):
    async def test(self):
        with self.assertRaises(exceptions.InvalidConfigurationException):
            super().setUp(max_file_size=1, max_buffer_size=2)
        with self.assertRaises(exceptions.InvalidConfigurationException):
            super().setUp(max_file_size=2**32, max_buffer_size=2)


class AioDBTestErrorInvalidGenesisBytes(AioDiskDBTestCase):
    async def test(self):
        with self.assertRaises(exceptions.InvalidConfigurationException):
            super().setUp(max_file_size=1, max_buffer_size=1, genesis_bytes=b'testtest')


class AioDBTestErrorDBRestartedAfterError(AioDiskDBTestCase):
    async def test(self):
        with self.assertRaises(exceptions.InvalidConfigurationException):
            super().setUp(max_file_size=1, max_buffer_size=1, genesis_bytes=b'testtest')


class AioDBTestErrorWrongFilesPrefix(AioDiskDBTestCase):
    async def test(self):
        with self.assertRaises(exceptions.InvalidConfigurationException):
            super().setUp(file_prefix='1222')
        with self.assertRaises(exceptions.InvalidConfigurationException):
            super().setUp(file_prefix='__aaa')
        with self.assertRaises(exceptions.InvalidConfigurationException):
            super().setUp(file_prefix='')

    def tearDown(self) -> None:
        pass


class AioDBTestStopTimeoutError(AioDiskDBTestCase):
    def setUp(self, *a, **kw):
        super().setUp(max_file_size=1, max_buffer_size=1, timeout=1)

    async def _fail_flush(self):
        try:
            await self.sut._flush_buffer()
        except exceptions.NotRunningException as e:
            print('Expected exception raised:', str(e))

    async def test(self):
        await self._run(expect_failure='wrong state, cannot recover. buffer lost.')
        await self.sut.add(b'a')
        await self.sut._write_lock.acquire()
        self.loop.create_task(self._fail_flush())
        with self.assertRaises(exceptions.TimeoutException):
            await self.sut.stop()
        self.sut._write_lock.release()

    def tearDown(self) -> None:
        pass


class AioDBTestFileSizeChangedError(AioDiskDBTestCase):
    def setUp(self, *a, **kw):
        super().setUp(max_file_size=10, max_buffer_size=2, timeout=1)

    async def _fail_flush(self):
        try:
            await self.sut._flush_buffer()
        except exceptions.InvalidDataFileException as e:
            print('Expected exception raised:', str(e))

    async def test(self):
        await self._run(expect_failure='wrong state, cannot recover. buffer lost.')
        await self.sut.add(os.urandom(2 * 1024**2))
        await self.sut._flush_buffer()
        filename = str(self.sut.path) + '/data00000.dat'

        with open(filename, 'r+b') as f:
            f.seek(os.path.getsize(filename)-10)
            f.truncate()
        await self.sut.add(os.urandom(2 * 1024 ** 2))
        await self._fail_flush()

    def tearDown(self) -> None:
        pass


class AioDBTestMissingFileException(AioDiskDBTestCase):
    def setUp(self, *a, **kw):
        super().setUp(max_file_size=10, max_buffer_size=2, timeout=1)

    async def _fail_flush(self):
        try:
            await self.sut._flush_buffer()
        except exceptions.FilesInconsistencyException as e:
            print('Expected exception raised:', str(e))

    async def test(self):
        await self._run(expect_failure='wrong state, cannot recover. buffer lost.')
        await self.sut.add(os.urandom(2 * 1024**2))
        await self.sut._flush_buffer()
        filename = str(self.sut.path) + '/data00000.dat'
        os.remove(filename)
        await self.sut.add(os.urandom(2 * 1024 ** 2))
        await self._fail_flush()

    def tearDown(self) -> None:
        pass



import abc
import asyncio
import time

import typing

from aiodiskdb import exceptions
from aiodiskdb.internals import ensure_running, GracefulExit
from aiodiskdb.local_types import EventsHandlers, ItemLocation


class AsyncLockable(metaclass=abc.ABCMeta):
    def __init__(self, *_, **__):
        super().__init__()
        self._write_lock = asyncio.Lock()
        self._transaction_lock = asyncio.Lock()
        self._read_lock = asyncio.Lock()
        self._reads_count = 0

    def _incr_read(self):
        self._reads_count += 1

    def _decr_read(self):
        self._reads_count -= 1


class AsyncObservable(metaclass=abc.ABCMeta):
    def __init__(self, *_, **__):
        super().__init__(self, *_, **__)
        self._events = EventsHandlers()

    @property
    def events(self):
        return self._events


class AsyncRunnable(AsyncObservable, AsyncLockable, metaclass=abc.ABCMeta):
    def __init__(self, *_, timeout=0, **__):
        super().__init__(*_, **__)
        self._running = False
        self._error = False
        self._set_stop = False
        self._do_stop = False
        self._timeout = timeout
        self._blocking_stop = False

    @abc.abstractmethod
    def _pre_stop_signal(self):
        pass  # pragma: no cover

    def on_stop_signal(self):
        """
        Non async method. Handle stop signals.
        """
        if self._pre_stop_signal():
            raise GracefulExit()

    @abc.abstractmethod
    async def _pre_loop(self):
        pass  # pragma: no cover

    @abc.abstractmethod
    async def _run_loop(self):
        pass  # pragma: no cover

    @abc.abstractmethod
    def _teardown(self):
        pass  # pragma: no cover

    @property
    def running(self):
        return self._running

    @ensure_running(False)
    async def run(self):
        """
        Must be launched before using the Database as a non blocking task.
        example:
        loop.create_task(instance.run())
        loop.run_until_complete()
        """
        loop = asyncio.get_event_loop()
        try:
            await self._pre_loop()
        except Exception as e:
            self._running = False
            self._error = e
            raise
        start_fired = False
        while 1:
            if not start_fired and self.running and self.events.on_start:
                loop.create_task(self.events.on_start(time.time()))
                start_fired = True
            if self._blocking_stop:
                break
            try:
                if self._set_stop:
                    await self._teardown()
                    self._do_stop = True
                    break
                await self._run_loop()
                await asyncio.sleep(0.005)
                self._running = True
            except Exception as e:
                if self.events.on_failure:
                    loop.create_task(self.events.on_failure(time.time()))
                self._running = False
                self._error = e
                self.events.on_stop and \
                    loop.create_task(self.events.on_stop(time.time()))
                raise
        self.events.on_stop and loop.create_task(self.events.on_stop(time.time()))
        self._running = False

    @ensure_running(True)
    async def stop(self):
        stop_requested_at = time.time()
        self._set_stop = True
        while time.time() - stop_requested_at < self._timeout:
            if not self._running:
                return True
            await asyncio.sleep(0.1)
        raise exceptions.TimeoutException(f'Loop is still running after {self._timeout} seconds')


class AioDiskDBTransactionAbstract(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def add(self, data: bytes) -> None:
        pass  # pragma: no cover

    @abc.abstractmethod
    def commit(self) -> typing.Iterable[ItemLocation]:
        pass  # pragma: no cover



import asyncio
import os
import random
import time

from aiodiskdb import exceptions
from test import AioDiskDBTestCase, run_test_db, AioDiskDBConcurrencyTest


class TestAioDiskDBTransaction(AioDiskDBTestCase):

    @run_test_db
    async def test(self):
        location1 = await self.sut.add(b'data1')
        transaction = await self.sut.transaction()
        transaction.add(b'data2')
        transaction.add(b'data3')
        await transaction.commit()
        with open(self._path + '/data00000.dat', 'rb') as f:
            x = f.read()
        self.assertEqual(x, self.sut._bake_new_file_header() + b'data1data2data3')
        location1.size += 10  # increase the location size to read contiguous data
        self.assertEqual(
            b'data1data2data3',
            await self.sut.read(location1)
        )


class TestAioDiskDBConcurrentTransactions(AioDiskDBConcurrencyTest):
    @run_test_db
    async def test(self):
        self._running_test = True
        self.loop.create_task(self._random_reads())
        data_stored = dict()
        total_size = 0
        s = time.time()
        sizes = []
        while sum(data_stored.values()) < 1000 * 1024 ** 2:
            random_transaction = random.randint(0, 10)
            if random_transaction < 3:
                """
                Add some data by transactions
                """
                transaction = await self.sut.transaction()
                transactions_data = []
                for x in range(0, random.randint(1, 10)):
                    size = random.randint(1024, 1024 ** 2)
                    sizes.append(size)
                    data = os.urandom(size)
                    transactions_data.append(data)
                    transaction.add(data)
                    await asyncio.sleep(0.001)
                locations = await transaction.commit()
                self._writes_count += 1
                self._transactions += 1
                for x in zip(locations, transactions_data):
                    self._data.append(x)  # append [location, data] so the random reads can request it
                    tx_chunk_size = len(x[1])
                    data_stored.setdefault(x[0].index, 0)
                    data_stored[x[0].index] += tx_chunk_size
                    total_size += tx_chunk_size
            else:
                """
                Mix normal adds
                """
                size = random.randint(1024, 1024 ** 2)
                sizes.append(size)
                data = os.urandom(size)
                location = await self.sut.add(data)
                self._writes_count += 1
                data_stored.setdefault(location.index, 0)
                self._data.append([location, data])
                data_stored[location.index] += size
                total_size += size
                await asyncio.sleep(0.00001)
                self.assertTrue(self._ongoing_reads, msg='reads failed')
        self._pause_reads = True
        print(f'R/W concurrency test over. Duration: {time.time() - s:.2f}s, '
              f'Reads: {self._reads_count}, Writes: {self._writes_count}, Transactions: {self._transactions}'
              f'Bandwidth: {total_size // 1024 ** 2}MB, '
              f'Avg file size: {sum(sizes) / len(sizes) // 1024}kB'
              )
        current_reads = self._reads_count

        # test is over, repeat the random reads with a new DB instance on the same data.
        print('Read only test with no-cache instance:')
        while self._ongoing_reads:
            await asyncio.sleep(0.2)
        await self.sut.stop()
        self._setup_sut()
        await self._run()
        while not self.sut.running:
            await asyncio.sleep(0.01)
        self._pause_reads = False
        s = time.time()
        while time.time() - s < 10:
            await asyncio.sleep(2)
            self.assertTrue(self._ongoing_reads, msg='reads failed')
        self._stop_reads = True
        print(f'Read only test from disk over. Reads: {self._reads_count - current_reads}')


class TransactionErrors(AioDiskDBTestCase):
    def setUp(self, *a):
        super().setUp(*a, max_file_size=2, max_buffer_size=1)

    @run_test_db
    async def test(self):
        transaction = await self.sut.transaction()
        with self.assertRaises(exceptions.WriteFailedException):
            transaction.add(os.urandom(3*1024**2))
        with self.assertRaises(exceptions.EmptyTransactionException):
            await transaction.commit()
        transaction.add(b'test')
        await transaction.commit()
        with self.assertRaises(exceptions.TransactionAlreadyCommittedException):
            await transaction.commit()
        with self.assertRaises(exceptions.TransactionAlreadyCommittedException):
            await transaction.add(b'bbb')
        transaction = await self.sut.transaction()
        for x in range(0, 100):
            transaction.add(os.urandom(1024**2))
        self.assertTrue(self.sut.running)
        self.loop.create_task(transaction.commit())
        await asyncio.sleep(0.01)
        self.assertTrue(self.sut.running)
        with self.assertRaises(exceptions.TransactionCommitOnGoingException):
            transaction.add(b'aaaa')



import os
from aiodiskdb import exceptions
from aiodiskdb.local_types import ItemLocation, WriteEvent
from test import AioDiskDBTestCase, run_test_db


class TestLTRIM(AioDiskDBTestCase):
    @run_test_db
    async def test(self):
        item_location = await self.sut.add(b'aaaa_1')
        self.assertEqual(
            ItemLocation(0, 0, 6),
            item_location
        )
        item_location_2 = await self.sut.add(b'bbbb_2')
        self.assertEqual(
            ItemLocation(0, 6, 6),
            item_location_2
        )
        item_location_3 = await self.sut.add(b'cccc_3')
        self.assertEqual(
            ItemLocation(0, 12, 6),
            item_location_3
        )
        await self.sut.ltrim(0, 6, safety_check=b'_1')
        item_2 = await self.sut.read(item_location_2)
        self.assertEqual(item_2, b'bbbb_2')
        item_3 = await self.sut.read(item_location_3)
        self.assertEqual(item_3, b'cccc_3')
        with self.assertRaises(exceptions.InvalidTrimCommandException):
            await self.sut.ltrim(0, 6, safety_check=b'_1')
        await self.sut.ltrim(0, 12, safety_check=b'_2')
        item_2 = await self.sut.read(item_location_2)
        self.assertIsNone(item_2)
        item_3 = await self.sut.read(item_location_3)
        self.assertEqual(item_3, b'cccc_3')



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='aiodiskdb',
    version='0.2',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/mempoolco/aiodiskdb/',
    license='MIT',
    author='Guido Dassori',
    author_email='guido.dassori@gmail.com',
    python_requires='>=3.8',
    description='Embeddable minimal asynchronous on disk DB',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    include_package_data=True,
    packages=['aiodiskdb'],
    package_dir={'aiodiskdb': 'aiodiskdb'},
)



import asyncio
from unittest import IsolatedAsyncioTestCase

from aiodiskdb import exceptions
from aiodiskdb.aiodiskdb import AioDiskDB
from aiodiskdb.local_types import EventsHandlers


class TestEventsHandlerStrictTyping(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.sut = EventsHandlers()
        self._dummy_ex = 0

    async def _dummy(self, *a, **kw):
        self._dummy_ex += 1

    async def test(self):
        with self.assertRaises(TypeError):
            self.sut.on_start = lambda w: ''

        self.sut.on_start = self._dummy
        await self.sut.on_start()
        self.sut.on_start = None
        self.assertEqual(1, self._dummy_ex)


class AioDiskDBTestCase(IsolatedAsyncioTestCase):
    _path = '/tmp/aiodiskdb_test'

    def setUp(self, max_file_size=128, max_buffer_size=16, overwrite=True):
        self.loop = asyncio.get_event_loop()
        self._overwrite = True
        self._max_file_size = max_file_size
        self._max_buffer_size = max_buffer_size
        self._setup_sut()
        self.sut.destroy_db()
        self._overwrite = overwrite
        self._setup_sut()

    def tearDown(self) -> None:
        self.sut.destroy_db()

    def _setup_sut(self):
        self.sut = AioDiskDB(
            self._path,
            create_if_not_exists=True,
            read_timeout=5,
            max_file_size=self._max_file_size,
            max_buffer_size=self._max_buffer_size,
            overwrite=self._overwrite
        )


def run_test_db(f):
    async def _decorator(self, *a, **kw):
        try:
            self.loop.create_task(self.sut.run(), name='aiodiskdb_main_loop')
            while not self.sut.running:
                await asyncio.sleep(0.01)
            return await f(self, *a, **kw)
        finally:
            try:
                await self.sut.stop()
            except exceptions.NotRunningException:
                print('run_test_db requested to shutdown a not running database')
    return _decorator
